/*  Copyright (C) 1998-9  Richard Braakman, Anthony Towns
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <assert.h>

#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/wait.h>

#include "run-scripts.h"

static char *mytime(void) {
    if (verbose) {
        static char buffer[100] = "<";
        time_t t = time(NULL);
        struct tm *local = localtime(&t);

        strftime( buffer+1, 99, "%c", local );
        strcat( buffer, "> ");
        return buffer;
    } else {
        return "";
    }
}   
    

struct script {
    /* set by run_scripts() */
    char *filename;

    /* set by parse_fileheader() */
    char *name;
    int level;
    struct stringset *needs;
    struct stringset *provides;
    
    /* set by dependency calculation */
    struct script **depends;
    int waiting_deps; /* number of dependencies yet to finish */

    /* set at execution */
    struct script *next;
    int completed;
    char *linebuffer;
    int running; /* if non-zero, following fields are valid */
    pid_t pid;
    int out_fd;
#if 0
    int err_fd;
#endif
};
typedef struct script script;

static script *scripts = NULL;
int n_scripts = 0;

/* depends[n][0...k-1] 
 *          are the indices of scripts that depend on script n
 * depends[n][k] = -1;
 */

int get_word(char **buf) {
    int len;
    
    while (isspace(**buf) || **buf == ',') (*buf)++;
    if (!**buf) return 0;
 
    len = 0;

    while (!isspace((*buf)[len]) && (*buf)[len] != ',' && (*buf)[len]) len++;
    
    return len;
}

char *skip_fieldname(char *buffer, char *name) {
    while (*buffer && isspace(*buffer)) buffer++;
    
    if (strncmp(buffer, name, strlen(name)) != 0) return NULL;
    
    buffer += strlen(name);
    if (*buffer != ':') return NULL;
    
    return ++buffer;
}

int parse_string(char *buffer, char *name, char **string) {
    int len;
    buffer = skip_fieldname(buffer, name);
    if (buffer == NULL) return 0;
    
    len = get_word(&buffer);
    if (len == 0) return 0;
    
    *string = malloc(len + 1);
    strncpy(*string, buffer, len);
    (*string)[len] = '\0';
    
    return 1;
}

int parse_strings(char *buffer, char *name, struct stringset **strings) {
    int len;
    buffer = skip_fieldname(buffer, name);
    if (buffer == NULL) return 0;

    if (debug) fprintf(stderr, "[[ %s: ", name);
    while ((len = get_word(&buffer))) {
	if (debug) fprintf(stderr, "(%d,%s) ", len, buffer);
	stringset_addelem(*strings, buffer, len);
	buffer += len;
    }
    if (debug) fprintf(stderr, "\n");
    
    return 1;
}

int parse_fileheader(script *script) {
    FILE *file;
    char buffer[1000];
    char *pch;
    
    script->name = NULL;
    script->level = -1;
    script->needs = NULL;
    script->provides = NULL;
    
    file = fopen(script->filename, "r");
    if (!file) return 0; /* invalid file */
    
    while (fgets(buffer, 1000, file)) {
	if (buffer[strlen(buffer)-1] == '\n') { /* get rid of trailing \n */
	    buffer[strlen(buffer)-1] = '\0';
	}

	if (buffer[0] != '#') break;
	
	if (buffer[1] == '!') continue;
	
	if (!script->needs && !script->provides) {
	    if (parse_string(buffer+1, "Name", &script->name)) {
		script->needs = stringset_new();
		script->provides = stringset_new();
	    }
	    continue;
	}
	
	if (parse_strings(buffer+1, "Needs", &script->needs))
	    continue;
	if (parse_strings(buffer+1, "Provides", &script->provides))
	    continue;
    }
    
    fclose(file);
	
    /* attempt to determine name from filename */
    pch = strrchr(script->filename, '/');
    if (pch) pch++;
    else     pch = script->filename;
    
    if ((pch[0] == 'S' || pch[0] == 'K')
	 && isdigit(pch[1]) && isdigit(pch[2])) 
    {
	script->level = 10 * (pch[1] - '0') + (pch[2] - '0');
	if (!script->name) script->name = pch + 3;
    } else {
	script->level = -1;
	if (!script->name) script->name = pch;
    }
    
    return 1;
}    

void add_dependency(int from, int to) {
    int x;

    if (debug) fprintf( stderr, "%s depends upon %s, apparently\n", 
	     scripts[from].name, scripts[to].name );
    
    for (x = 0; scripts[to].depends[x] != NULL; x++) {
	if (scripts[to].depends[x] == &scripts[from]) return;
    }

    scripts[to].depends[x++] = &scripts[from];
    scripts[to].depends[x] = NULL;
    
    scripts[from].waiting_deps++;
}

int generate_dependencies(int reverse) {
    int i,j;
    int x,y;
    
    for (i = 0; i < n_scripts; i++) {
	scripts[i].waiting_deps = 0;
	scripts[i].depends = malloc(sizeof(int) * (n_scripts + 1));
	if (!scripts[i].depends) return 0;
	scripts[i].depends[0] = NULL;
    }
    
    for (i = 0; i < n_scripts; i++) {
	int foundall = 0;
	if (scripts[i].needs) {
	    foundall = 1;
	    for (x = 0; x < scripts[i].needs->count; x++) {
		int found = 0;
		for (j = 0; !found && j < n_scripts; j++) {
		    if (strcmp(scripts[i].needs->strings[x], 
				 scripts[j].name) == 0)
			found = 1;
		    for (y = 0; !found && y < scripts[j].provides->count; y++)
		    {
			if (strcmp(scripts[i].needs->strings[x],
				     scripts[j].provides->strings[y]) == 0)
			    found = 1;
		    }
		    if (found)
			if (reverse)
			    add_dependency(j,i);
			else
			    add_dependency(i,j);
		}
		if (!found) {
		    foundall = 0;
		    if (debug)
                        fprintf( stderr, "unfulfilled dep: %s -> \"%s\"\n",
			    scripts[i].name, scripts[i].needs->strings[x] );
		}
	    }
	}

	if (!foundall && scripts[i].level >= 0) {
	    for (j = 0; j < n_scripts; j++) {
		if (scripts[j].level >= 0 
		     && scripts[j].level < scripts[i].level) 
		{
		    add_dependency(i,j);
		}
	    }
	}
    }

    return 1;
}

void addlast(script **queue, script *el) {
    /* queue is the address of a pointer to the last element in the list
     * ie (*queue)->next is the first element. Elements are added to the
     * end of the list 
     */

    if (*queue) {
	el->next = (*queue)->next;
	(*queue)->next = el;
    } else {
	el->next = el;
    }
    *queue = el;
	
}

script *delfirst(script **queue) {
    /* el is an element in queue, possibly the only element */

    script *result = (*queue)->next;

    if (*queue == (*queue)->next) {
	*queue = NULL;
    } else {
	(*queue)->next = result->next;
    }
    
    return result;
}

int launch(script *q, char *arg) {
    int stdout_fd[2];
#if 0
    int stderr_fd[2];
#endif

    if (pipe(stdout_fd) != 0) return 0;
#if 0
    if (pipe(stderr_fd) != 0) return 0;
#endif

    switch(q->pid = fork()) {
    case -1: /* failed */
	return 0;
    case 0: /* child */
    {
	char *scriptname = "./fakeinit";
	char *pch;

	scriptname = q->filename;

	close(stdout_fd[0]); /* don't need to read from pipes */
	dup2(stdout_fd[1], 1); /* point stdout and stderr at the pipes */

#if 0
	close(stderr_fd[0]);
	dup2(stderr_fd[1], 2);
#endif

	if (no_act) {	
            printf("%s*Not* starting %s\n", mytime(), scriptname );
            exit(0);
	} else {
	    pch = scriptname + strlen(scriptname);
	    if (pch - scriptname > 3 && strcmp(pch - 3, ".sh") == 0) {
       	        execlp("/bin/sh", "/bin/sh", scriptname, arg, NULL);
	        perror(scriptname);
	        exit(0);
	    } else {
	        execlp(scriptname, scriptname, arg, NULL);
	        perror(scriptname);
	        exit(0);
	    }
        }
    }
    default: /* parent */
	assert(q->pid > 0);
	close(stdout_fd[1]); /* don't need to write to pipes */
	q->out_fd = stdout_fd[0];
#if 0
	close(stderr_fd[1]);
	q->err_fd = stderr_fd[0];
#endif

	return 1;
    }
}

void add_to_waiting(script **waiting, script *q) {
    addlast(waiting,q);
    q->completed = 1;
    q->linebuffer = malloc(1);
    q->linebuffer[0] = '\0';
}

/* variables */

script *running = NULL;
int n_running = 0;

script *waiting = NULL;

int toprocess = 0;

int dealwithio(fd_set *readfds) 
{
    char buffer[1001];
    int changed;
    int x;
    script *q;
    int j;
    int n;

    n = n_running;
    while (n > 0) {
	n--;
	q = running->next;
	
	if (FD_ISSET(q->out_fd, readfds)) {	    
	    x = read(q->out_fd, buffer, 1000);
	    
	    if (x > 0) {
		char *pch = buffer;
		char *pchnext;
		
		buffer[x] = '\0';
		
		while(*pch && (pchnext = strchr(pch, '\n'))) {
		    *pchnext = '\0';
		    printf("%s[%s] %s%s\n", mytime(), q->name, 
                                             q->linebuffer, pch);
		    q->linebuffer[0] = '\0';
		    pch = pchnext + 1;
		}
		if (*pch) {
		    q->linebuffer = realloc(q->linebuffer, 
					    strlen(q->linebuffer) 
					    + strlen(pch) + 1);
		    strcat(q->linebuffer, pch);
		}
	    }
	}

	if (q->pid == waitpid(q->pid, NULL, WNOHANG)) {
	    delfirst(&running);			
	    n_running--;
	    
	    if (verbose) 
	        printf("%sFinished %s\n", mytime(), q->name);
            

	    q->running = 0;
	    for (j = 0; q->depends[j]; j++) {
		assert(q->depends[j]->waiting_deps > 0);
		
		q->depends[j]->waiting_deps--;
		if (q->depends[j]->waiting_deps == 0) {
		    add_to_waiting(&waiting, q->depends[j]);
		}
	    }
	    toprocess--;
	    
	    q = running;
	    changed = 1;
	}

	running = q;
    }
    return changed;
}

int initfds(script *running, fd_set *readfds) {
    script *q;
    int max = -1;

    FD_ZERO(readfds);
    if (running) {
	q = running->next;
	do {
	    FD_SET(q->out_fd, readfds); 
	    if (max < q->out_fd) max = q->out_fd;

#if 0
	    FD_SET(q->err_fd, readfds); 
	    if (max < q->err_fd) max = q->err_fd;
#endif

	    q = q->next;
	} while(q != running->next);
    }

    return max;
}

int run_scripts(char **filenames, int count, char *argument, int reverse) {
    int i;
    static script empty;
    char *introstr = "Running"; /* default */

    if (argument != NULL) {
	if (strcmp(argument, "start") == 0) {
	    introstr = "Starting";
	} else if (strcmp(argument, "stop") == 0) {
	    introstr = "Stopping";
	}
    }

    scripts = (script *)malloc(count * sizeof(script));
    if (!scripts) {
	perror("malloc");
	return -1;
    }

    n_scripts = count;
    for (i = 0; i < count; i++) {
	scripts[i] = empty;
	scripts[i].filename = filenames[i];
	
	if (!parse_fileheader(&scripts[i])) {
	    free(scripts);
	    return -1;
	}
    }

    if (!generate_dependencies(reverse)) {
	return -1;
    }

    toprocess = count;    
    n_running = 0;
    running = NULL;
    waiting = NULL;
    
    while (toprocess > 0) {
	script *q;
	int first;

	/* We've got things to process, but we don't know which ones to
	 * start, and we're not running any. So just look for stuff, and
	 * hope. This is `last resort' code, essentially.
	 */
 
	if (running == NULL && waiting == NULL) {
	    script *least = NULL;

	    for (i = 0; i < count; i++) {
		if (!scripts[i].completed) {
		    if (scripts[i].waiting_deps == 0) {
			add_to_waiting(&waiting, &scripts[i]);
		    } else if (!least || least->waiting_deps 
			       > scripts[i].waiting_deps) 
		    {
			least = &scripts[i];
		    }
		}
	    }
	    if (!waiting) add_to_waiting(&waiting, least);
	}

	/* By now we should either be running stuff, or have a queue of
	 * things to run. If we've got stuff to start, start it. 
	 */
	first = 1;
	while (waiting) { /* && n_running < MAX_RUNNING */
	    q = delfirst(&waiting);

	    if (first) {
	        if (verbose) {
		    printf("%s%s %s", mytime(), introstr, q->name);
                }
		first = 0;
	    } else {
	        if (verbose) printf(", %s", q->name );
	    }
	    fflush(stdout);
	    if (q->waiting_deps != 0)
	        if (verbose) printf("[!]");

	    if ( launch(q, argument) ) {
		addlast(&running, q);
		n_running++;
		q->running = 1;
	    } else {
		toprocess--;
	    }
	}
	if (!first && verbose) printf("...\n"); fflush(stdout);
	
	if (running) {
	    fd_set readfds;
	    int max;
	    struct timeval tv;
	    do {
		tv.tv_sec = 1; tv.tv_usec = 0;
		max = initfds(running, &readfds);
		select(max+1, &readfds, NULL, NULL, &tv);
	    } while(!dealwithio(&readfds));
	}
    }

    return 0;
}
