/*  Copyright (C) 1998-9  Richard Braakman, Anthony Towns
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <errno.h>

#include "run-scripts.h"

/* These functions take pointers to 'struct stringset'. */
/* Originally I wanted NULL to be a valid representation of the empty
 * set, but that complicates the interface.  Let the caller handle it
 * if the caller wants to save space. */

static int error;

static void stringset_set_error(int err) {
    if (!error) error = err;
}

void stringset_clear_error(void) {
    error = 0;
}

int stringset_error(void) {
    return error;
}

static struct stringset *stringset_alloc(int prealloc) {
    struct stringset *set;

    set = (struct stringset *)malloc(sizeof(*set));
    if (!set) {
	stringset_set_error(errno);
	return NULL;
    }
    set->strings = NULL;
    set->count = 0;
    set->room = 0;

    if (prealloc > 0) {
	set->strings = (char **)malloc(prealloc * sizeof(char *));
	if (!set->strings) {
	    stringset_set_error(errno);
	    free(set);
	    return NULL;
	}
	set->room = prealloc;
    }
    return set;
}

static void stringset_widen(struct stringset *set) {
    char **newstrings;

    newstrings = realloc(set->strings, set->room * 2 * sizeof(*newstrings));
    if (newstrings) {
	set->strings = newstrings;
	set->room *= 2;
    } else {
	stringset_set_error(errno);
    }
}

/* Return 1 if the element was added, 0 if it was already present,
 * and -1 if there was an error. */
int stringset_addelem(struct stringset *set, char *elem, int len) {
    int i;

    if (len < 0)
	len = strlen(elem);

    for (i = 0; i < set->count; i++) {
	if (strncmp(set->strings[i], elem, len) == 0
	    && set->strings[i][len] == '\0')
	    return 0;
    }
    
    if (set->count == set->room) {
	stringset_widen(set);
	if (set->count == set->room)
	    return -1;
    }

    set->strings[set->count] = (char *)malloc(len);
    if (!set->strings[set->count]) {
	stringset_set_error(errno);
	return -1;
    }

    memcpy(set->strings[set->count], elem, len);
    set->strings[set->count][len] = '\0';
    set->count++;
    return 1;
}

int stringset_subelem(struct stringset *set, char *elem, int len) {
    int i;

    if (len < 0)
	len = strlen(elem);

    for (i = 0; i < set->count; i++) {
	if (strncmp(set->strings[i], elem, len) == 0
	    && set->strings[i][len] == '\0') {
	    free(set->strings[i]);
	    set->strings[i] = set->strings[set->count];
	    set->count--;
	    return 1;
	}
    }

    return 0;
}

/* Add all elements from set y to set x.  Return the actual number
 * of elements added (which may be < size(y) if there are overlaps),
 * or -1 in case of error. */
int stringset_add(struct stringset *x, struct stringset *y) {
    int i;
    int count;
    int res;

    /* Do it the easy way :-) */
    count = 0;
    for (i = 0; i < y->count; i++) {
	res = stringset_addelem(x, y->strings[i], -1);
	if (res < 0)
	    return -1;
	count += res;
    }
    return count;
}

#if 0
int stringset_sub(struct stringset *x, struct stringset *y) {
    int i;
    int count;
    int res;

    count = 0;
    for (i = 0; i < y->count; i++) {
	res = stringset_delelem(x, y->strings[i], -1);
	if (res < 0)
	    return -1;
	count += res;
    }
    return count;
}
#endif

/* Parse a comma-separated list of tokens, strip whitespace, and
 * construct a stringset from that. */
struct stringset *stringset_parse(char *line) {
    struct stringset *set;
    char *p;
    int len;
    int n;

    /* First find the number of elements */
    n = 1;
    for (p = line; *p; p++) {
	if (*p == ',')
	    n++;
    }

    /* Allocate space for that many */
    set = stringset_alloc(n);
    if (!set)
	return NULL;

    /* Fill in the strings */
    p = line;
    while (p && *p) {
	p += strspn(p, " \t\n");
	len = strcspn(p, " \t\n,");
	if (stringset_addelem(set, p, len) < 0)
	    return set;
	p = strchr(p + len, ',');
	if (p) p++;
    }

    return set;
}

/* Return 1 if x and y have no elements in common, 0 otherwise */
int stringset_empty_intersect(struct stringset *x, struct stringset *y) {
    int i, j;
    if (!x || !y)
	return 1;

    for (i = 0; i < x->count; i++) {
	for (j = 0; j < y->count; j++) {
	    if (strcmp(x->strings[i], y->strings[j]) == 0)
		return 0;
	}
    }

    return 1;
}

struct stringset *stringset_new(void) {
    return stringset_alloc(5);
}
